use uuid::Uuid;
use tokio::net::TcpListener;
use std::io::Cursor;

use crate::*;

#[tokio::test]
async fn test_data_transfer() -> Result<(), ProtocolError> {

    async fn server_thread() -> Result<(), ProtocolError> {
        let listener =
            TcpListener::bind("localhost:44447").await.or(Err(ProtocolError::StreamConnectError))?;

        while let Ok((stream, _)) = listener.accept().await {
            let mut stream = MCConnTcp::new(stream);

            stream.set_compression(Some(5));

            let mut packet = stream.read_packet().await?;

            let mut pack = Packet::empty(packet.id());

            pack.write_boolean(packet.read_boolean().await?).await?;
            pack.write_byte(packet.read_byte().await?).await?;
            pack.write_bytes(&packet.read_bytes(10).await?.clone()).await?;
            pack.write_double(packet.read_double().await?).await?;
            pack.write_float(packet.read_float().await?).await?;
            pack.write_i128_varint(packet.read_i128_varint().await?).await?;
            pack.write_u128_varint(packet.read_u128_varint().await?).await?;
            pack.write_int(packet.read_int().await?).await?;
            pack.write_long(packet.read_long().await?).await?;
            pack.write_short(packet.read_short().await?).await?;
            pack.write_uuid(&packet.read_uuid().await?).await?;
            pack.write_string(&packet.read_string().await?.clone()).await?;

            stream.write_packet(&pack).await?;

            stream.set_compression(None);

            let mut packet = stream.read_packet().await?;

            let mut pack = Packet::empty(packet.id());

            pack.write_boolean(packet.read_boolean().await?).await?;
            pack.write_byte(packet.read_byte().await?).await?;
            pack.write_bytes(&packet.read_bytes(10).await?.clone()).await?;
            pack.write_double(packet.read_double().await?).await?;
            pack.write_float(packet.read_float().await?).await?;
            pack.write_i128_varint(packet.read_i128_varint().await?).await?;
            pack.write_u128_varint(packet.read_u128_varint().await?).await?;
            pack.write_int(packet.read_int().await?).await?;
            pack.write_long(packet.read_long().await?).await?;
            pack.write_short(packet.read_short().await?).await?;
            pack.write_uuid(&packet.read_uuid().await?).await?;
            pack.write_string(&packet.read_string().await?.clone()).await?;

            stream.write_packet(&pack).await?;
        }

        Ok(())
    }

    tokio::spawn(async { let _ = server_thread().await; });
    
    let conn = MCConnTcp::connect("localhost:44447").await;

    while let Err(_) = conn {}

    let mut conn = conn?;

    conn.set_compression(Some(5));

    let mut pack = Packet::empty(0xfe);

    pack.write_boolean(true).await?;
    pack.write_byte(0x12).await?;
    pack.write_bytes(&vec![0x01, 0x56, 0x47, 0x48, 0xf5, 0xc2, 0x45, 0x98, 0xde, 0x99]).await?;
    pack.write_double(123456789.123456789f64).await?;
    pack.write_float(789456.44422f32).await?;
    pack.write_i128_varint(468927513325566).await?;
    pack.write_u128_varint(99859652365236523).await?;
    pack.write_int(77861346i32).await?;
    pack.write_long(789465123545678946i64).await?;
    pack.write_short(1233i16).await?;
    pack.write_uuid(&Uuid::try_parse("550e8400-e29b-41d4-a716-446655440000").map_err(|_| ProtocolError::CloneError)?).await?;
    pack.write_string("&packet.read_string()?").await?;

    conn.write_packet(&pack).await?;

    let mut packet = conn.read_packet().await?;

    assert_eq!(packet.read_boolean().await?, true);
    assert_eq!(packet.read_byte().await?, 0x12);
    assert_eq!(packet.read_bytes(10).await?, vec![0x01, 0x56, 0x47, 0x48, 0xf5, 0xc2, 0x45, 0x98, 0xde, 0x99]);
    assert_eq!(packet.read_double().await?, 123456789.123456789f64);
    assert_eq!(packet.read_float().await?, 789456.44422f32);
    assert_eq!(packet.read_i128_varint().await?, 468927513325566);
    assert_eq!(packet.read_u128_varint().await?, 99859652365236523);
    assert_eq!(packet.read_int().await?, 77861346i32);
    assert_eq!(packet.read_long().await?, 789465123545678946i64);
    assert_eq!(packet.read_short().await?, 1233i16);
    assert_eq!(packet.read_uuid().await?, Uuid::try_parse("550e8400-e29b-41d4-a716-446655440000").map_err(|_| ProtocolError::CloneError)?);
    assert_eq!(packet.read_string().await?, "&packet.read_string()?");

    conn.set_compression(None);

    let mut pack = Packet::empty(0xfe);
    
    pack.write_boolean(true).await?;
    pack.write_byte(0x12).await?;
    pack.write_bytes(&vec![0x01, 0x56, 0x47, 0x48, 0xf5, 0xc2, 0x45, 0x98, 0xde, 0x99]).await?;
    pack.write_double(123456789.123456789f64).await?;
    pack.write_float(789456.44422f32).await?;
    pack.write_i128_varint(468927513325566).await?;
    pack.write_u128_varint(99859652365236523).await?;
    pack.write_int(77861346i32).await?;
    pack.write_long(789465123545678946i64).await?;
    pack.write_short(1233i16).await?;
    pack.write_uuid(&Uuid::try_parse("550e8400-e29b-41d4-a716-446655440000").map_err(|_| ProtocolError::CloneError)?).await?;
    pack.write_string("&packet.read_string()?").await?;

    conn.write_packet(&pack).await?;

    let mut packet = conn.read_packet().await?;

    assert_eq!(packet.read_boolean().await?, true);
    assert_eq!(packet.read_byte().await?, 0x12);
    assert_eq!(packet.read_bytes(10).await?, vec![0x01, 0x56, 0x47, 0x48, 0xf5, 0xc2, 0x45, 0x98, 0xde, 0x99]);
    assert_eq!(packet.read_double().await?, 123456789.123456789f64);
    assert_eq!(packet.read_float().await?, 789456.44422f32);
    assert_eq!(packet.read_i128_varint().await?, 468927513325566);
    assert_eq!(packet.read_u128_varint().await?, 99859652365236523);
    assert_eq!(packet.read_int().await?, 77861346i32);
    assert_eq!(packet.read_long().await?, 789465123545678946i64);
    assert_eq!(packet.read_short().await?, 1233i16);
    assert_eq!(packet.read_uuid().await?, Uuid::try_parse("550e8400-e29b-41d4-a716-446655440000").map_err(|_| ProtocolError::CloneError)?);
    assert_eq!(packet.read_string().await?, "&packet.read_string()?");

    Ok(())
}

#[tokio::test]
async fn test_compression() -> Result<(), ProtocolError> {
    let mut conn = MCConn::new(Cursor::new(Vec::new()));
    conn.set_compression(Some(5));

    let mut packet_1 = Packet::empty(0x12);
    packet_1.write_bytes(b"1234567890qwertyuiopasdfghjklzxcvbnm").await?;
    dbg!(&packet_1);
    conn.write_packet(&packet_1).await?;

    conn.get_mut().set_position(0);

    let mut packet_2 = conn.read_packet().await?;

    assert_eq!(
        packet_2.read_bytes(36).await?,
        b"1234567890qwertyuiopasdfghjklzxcvbnm"
    );

    Ok(())
}
